/**
* Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
*
* Licensed under the Amazon Software License (the "License"). You may not use this file except in # compliance with the License.
* A copy of the License is located at http://aws.amazon.com/asl/
*
* or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS, # WITHOUT WARRANTIES OR CONDITIONS
* OF ANY KIND, express or implied. See the License for the specific # language governing permissions and limitations under the License.
**/

// This file includes sample lambda code for a Video Skills API integration on Fire TV, including responses for Discovery,
// SearchAndPlay, FastForward, Channel Change, and other directive types. Partners should use this code as a starting
// point, but then add their own logic. Partners should test and validate any code before using it in production.
// Amazon's recommended approach is to take directives received by the lambda and send them directly to the app via ADM.
// Any further processing can occur within the app.

// NOTE: Responses back to Alexa should NOT be modified, only code to process/send directives.

// NOTE: You should send directives from the lambda to your app, but you should NOT try to send messages from your app back
// to the lambda. Success responses from the lambda are only meant to indicate that the message was received by the lambda,
// not the app.

// NOTE: If your Lambda is timing out, please increase the timeout duration under Advanced settings in the configuration tab.

// NOTE: Although ADM can send messages from the Lambda to the app, there is no reliable mechanism for a lambda to receive
// messages from the app back into the lambda. Partners should not try to send messages from their app back to the lambda.
//Make sure to replace the values for CLIENT_ID (line 32), CLIENT_SECRET (line 33), and admRegistrationId (Line 298)

const AWS = require('aws-sdk');
const REQUEST = require('request');

const CLIENT_ID = '<enter client ID>';
const CLIENT_SECRET = '<enter client secret>';

exports.handler = (event, context, callback) => {
  console.log("Lambda was invoked by Alexa");
  var name = event.directive.header.name;
  console.log("Context: " + JSON.stringify(context, null, 2));
  console.log("Alexa Request: ", name, JSON.stringify(event));
  let directive = event.directive;
  if (directive) {

    if (directive.header.name == "Discover") {

      let resp = {
        event: {
          header: {
            messageId: directive.header.messageId,
            name: "Discover.Response",
            namespace: "Alexa.Discovery",
            payloadVersion: "4"
          },
          "payload": {
            "endpoints": [{
              "capabilities": [{
                  "interface": "Alexa.RemoteVideoPlayer",
                  "type": "AlexaInterface",
                  "version": "1.0"
                },
                {
                  "interface": "Alexa.PlaybackController",
                  "type": "AlexaInterface",
                  "version": "1.0"
                },
                {
                  "interface": "Alexa.SeekController",
                  "type": "AlexaInterface",
                  "version": "1.0"
                },
                {
                  "interface": "Alexa.ChannelController",
                  "type": "AlexaInterface",
                  "version": "1.0"
                }
              ],
              "endpointId": "VSKTV",
              "description": "VSKTV is the default device used for video skills developed for Fire TV integrations",
              "friendlyName": "VSKTV",
              "manufacturerName": "Amazon"
            }]
          }
        }
      };

      // Process the directive by sending the message to app
      sendMessageToDevice(JSON.stringify(event), context).then(
        function(body) {
          console.log("Sending ", directive.header.name, " response back to Alexa: ", JSON.stringify(resp));
          return callback(null, resp);
        },
        function(err) {
          console.log('Could not send the message to the device: ' + err);
          return callback(null);
        }
      )

    } else if (directive.header.name == "SearchAndPlay" ||
      directive.header.name == "SearchAndDisplayResults" ||
      directive.header.name == "Play" ||
      directive.header.name == "Resume" ||
      directive.header.name == "Pause" ||
      directive.header.name == "Next" ||
      directive.header.name == "Previous" ||
      directive.header.name == "Rewind" ||
      directive.header.name == "FastForward") {

      // Partner-specific logic to handle the directives goes here

      // Below, we send a "success" response back to Alexa, indicating the directive was received by the lambda
      let resp = {
        event: {
          "header": {
            "messageId": "not-required",
            "correlationToken": "not-required",
            "name": "Response",
            "namespace": "Alexa",
            "payloadVersion": "3"
          },
          "endpoint": {
            "scope": {
              "type": "DirectedUserId",
              "directedUserId": "not-required"
            },
            "endpointId": "not-required"
          },
          "payload": {}
        }
      };

      // Process the directive by sending the message to app
      sendMessageToDevice(JSON.stringify(event), context).then(
        function(body) {
          console.log("Sending ", directive.header.name, " response back to Alexa: ", JSON.stringify(resp));
          return callback(null, resp);
        },
        function(err) {
          console.log('Could not send the message to the device: ' + err);
          return callback(null);
        }
      )

    } else if (directive.header.name == "AdjustSeekPosition") {

      // Partner-specific logic to handle the AdjustSeekPosition directive goes here

      // Below, we send a "success" response back to Alexa, indicating the directive was received by the lambda
      let resp = {
        event: {
          "header": {
            "messageId": "not-required",
            "correlationToken": "not-required",
            "name": "StateReport",
            "namespace": "Alexa.SeekController",
            "payloadVersion": "3"
          },
          "endpoint": {
            "scope": {
              "type": "DirectedUserId",
              "directedUserId": "not-required"
            },
            "endpointId": "not-required"
          },
          "payload": {
            "properties": [{
              "name": "positionMilliseconds",
              "value": 0
            }]
          }
        }
      };

      // Process the directive by sending the message to app
      sendMessageToDevice(JSON.stringify(event), context).then(
        function(body) {
          console.log("Sending ", directive.header.name, " response back to Alexa: ", JSON.stringify(resp));
          return callback(null, resp);
        },
        function(err) {
          console.log('Could not send the message to the device: ' + err);
          return callback(null);
        }
      )

    } else if (directive.header.name == "ChangeChannel") {

      // Partner-specific logic to handle the ChangeChannel directive goes here
      // Code for sending this directive from the lambda to the app goes here

      // Below, we send a "success" response back to Alexa, indicating the directive was received by the lambda
      let resp = {
        context: {
          "properties": [{
            "namespace": "Alexa.ChannelController",
            "name": "channel",
            "value": {
              "number": "1234",
              "callSign": "not-required",
              "affiliateCallSign": "not-required"
            },
            "timeOfSample": "2017-02-03T16:20:50.52Z",
            "uncertaintyInMilliseconds": 0
          }]
        },
        event: {
          "header": {
            "messageId": "not-required",
            "correlationToken": "not-required",
            "name": "Response",
            "namespace": "Alexa",
            "payloadVersion": "3"
          },
          "endpoint": {
            "endpointId": "not-required"
          },
          "payload": {}
        }
      };

      // Process the directive by sending the message to app
      sendMessageToDevice(JSON.stringify(event), context).then(
        function(body) {
          console.log("Sending ", directive.header.name, " response back to Alexa: ", JSON.stringify(resp));
          return callback(null, resp);
        },
        function(err) {
          console.log('Could not send the message to the device: ' + err);
          return callback(null);
        }
      )

    } else if (directive.header.name == "AcceptGrant")
    {
        console.log(directive.header.name, " directive received from Alexa");
        let resp = {
            "event": {
                "header": {
                    "messageId": "abc-123-def-456",
                    "namespace": "Alexa.Authorization",
                    "name": "AcceptGrant.Response",
                    "payloadVersion": "3"
                    },
            "payload": {}
          }
        };
        console.log("Sending ", directive.header.name, " response back to Alexa: ", JSON.stringify(resp));
        callback(null, resp);

    } else {
      console.log("Directive is not supported. Ignoring it.");
      callback(null, 'Directive is not supported. Ignoring it.');
    }
  } else {
    console.log('Directive is null. Ignoring it.');
    callback(null, 'Directive is null. Ignoring it.');
  }
};



// This function provides sample code to send a directive from the lambda to a Fire TV app via ADM. Partners should
// combine the sample code in this function with their own client id, client secret, etc
// to create a full implementation.
//
// Prerequisites:
//
// 1) Understand Amazon Device Messaging (ADM),which lets you send messages to Amazon devices that run your app.
// https://developer.amazon.com/public/apis/engage/device-messaging/tech-docs/adm-understanding
//
// 2) A FireTV application that is integrated with ADM.
// https://developer.amazon.com/public/apis/engage/device-messaging/tech-docs/adm-setting-up
function fetchAccessToken(event, context) {
  const ADM_AUTH_URL = 'https://api.amazon.com/auth/O2/token';
  var responseToAlexa = 'This is a dummy response. Replace it with your own value.';

  return new Promise(function(resolve, reject) {

    // Here, we get the accessToken from ADM using your credentials. If it succeeds, we call the sendMessageToDevice()
    // function. We use the "request" module for sending the request. ADM documentation regarding request headers:
    // https://developer.amazon.com/public/apis/engage/device-messaging/tech-docs/adm-requesting-access-token
    REQUEST({
      url: ADM_AUTH_URL,
      method: 'POST',
      auth: {
        user: CLIENT_ID,
        pass: CLIENT_SECRET
      },
      form: {
        'grant_type': 'client_credentials',
        'scope': 'messaging:push'
      }
    }, function(err, res) {
      if (err || res.statusCode != 200) {
        console.log("Unable to retrieve access token: " + err + " Response code: " + res.statusCode);
        reject("No access token!");
      } else {
        console.log('Got Access Token Type:' + JSON.parse(res.body).token_type + ' Expires In:' + JSON.parse(res.body).expires_in);
        resolve(JSON.parse(res.body).access_token);
      }
    });
  });

}


// This function fetches an access token before sending a message to your app via ADM. We reference your app instance
// via the ADM Registration ID. In production, partners should read  ADM Registration IDs from directives, and
// dynamically route messages to the correct app, rather than hardcoding.

// This method is a blocking method and will not respond back to Alexa till it gets a response back from the app.
// If response time becomes an issue try sending message to device in async mode
function sendMessageToDevice(message, context) {
  const UTIL = require('util');
  const ADM_MESSAGE_URL = 'https://api.amazon.com/messaging/registrations/%s/messages';
  var admRegistrationId = JSON.parse(message).directive.endpoint.cookie.applicationInstanceId;
  console.log("ADM Registration ID: " + admRegistrationId);

  return fetchAccessToken().then((accessToken) => {

    // Construct Request Headers per ADM specification:
    // https://developer.amazon.com/public/apis/engage/device-messaging/tech-docs/adm-sending-message
    var postData = {
      data: {
        'message': message,
        'timeStamp': new Date().valueOf().toString()
      },
      consolidationKey: 'ADM_Enqueue_Sample',
      expiresAfter: 3600
    }
    var url = UTIL.format(ADM_MESSAGE_URL, admRegistrationId);
    var options = {
      method: 'post',
      body: postData,
      json: true,
      url: url,
      auth: {
        'bearer': accessToken
      },
      headers: {
        'content-Type': 'application/json',
        'accept': 'application/json',
        'X-Amzn-Type-Version': 'com.amazon.device.messaging.ADMMessage@1.0',
        'X-Amzn-Accept-Type': 'com.amazon.device.messaging.ADMSendResult@1.0'
      }
    };

    return new Promise(function(resolve, reject) {
      console.log('Calling device...');
      // Use "request" module to send the request to ADM-server
      REQUEST(options, function (err, res, body) {
          if (err || res.statusCode != 200) {
              console.log('Error sending message to the app: ' + err + ' Response code: ' + res.statusCode);
              console.log('Response: ' + JSON.stringify(body));
              reject(null);
          } else {
              console.log('Response from the app: ' + JSON.stringify(body));
              resolve(body);
          }
      })

//      resolve('Success');

    });

  });


}
